# Version 0.1
#
#  GeologicTimeScale is copyright 2005 Jeffrey G. Pittman.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA

from Products.Archetypes.public import BaseFolderSchema, Schema
from Products.Archetypes.public import StringField, IntegerField, FloatField, LinesField, BooleanField
from Products.Archetypes.public import LinesWidget, TextAreaWidget, IdWidget, IntegerWidget, DecimalWidget, StringWidget, BooleanWidget, MultiSelectionWidget
from Products.Archetypes.public import BaseFolder, registerType
from Products.GeologicTimeScale.config import PROJECTNAME
from Products.CMFCore import CMFCorePermissions

import sys,string,cStringIO

from config import *

from AccessControl import ClassSecurityInfo, ModuleSecurityInfo

try :
    from Shared.reportlab.platypus.flowables import Image
    from Shared.reportlab.platypus.paragraph import Paragraph
    from Shared.reportlab.platypus.doctemplate import *
    from Shared.reportlab.lib.units import inch
    from Shared.reportlab.lib.units import cm
    from Shared.reportlab.lib import styles
    from Shared.PIL import Image as PIL_Image
    from reportlab.lib import colors as rlcolors
    from reportlab.graphics.shapes import *
    from reportlab.graphics.charts.textlabels import Label
    from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, Paragraph
    from reportlab.lib.styles import ParagraphStyle
    from reportlab.lib.pagesizes import A4
except ImportError :
    from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, Paragraph
    from reportlab.lib.styles import ParagraphStyle
    from reportlab.lib.pagesizes import A4
    from reportlab.platypus.flowables import Image
    from reportlab.platypus.paragraph import Paragraph
    from reportlab.platypus.doctemplate import *
    from reportlab.lib.units import inch
    from reportlab.lib.units import cm
    from reportlab.lib import styles
    from PIL import Image as PIL_Image
    from reportlab.lib import colors as rlcolors
    from reportlab.graphics.shapes import *
    from reportlab.graphics.charts.textlabels import Label

from Products.GeologicTimeScale.drawing import GeologicTimeScaleDrawing
from Products.GeologicTimeScale.sample_data import sample_point_events
from Products.GeologicTimeScale.sample_data import sample_range_events

# The following modified from rlzopecatalog.py, a modification of rlzope.py
class DrawingPDFDoc :
    class DrawingPageTemplate(PageTemplate) :
        def __init__(self, parent) :
            self.parent = parent
            #content = Frame(0.75 * inch, 0.5 * inch, parent.document.pagesize[0] - 1.25 * inch, parent.document.pagesize[1] - (1.5 * inch))
            #content = Frame(0.5*cm, 0.5*cm, 17*cm, 27*cm, id='F1')
            content = Frame(0.0, 0.0, 800, 4700, id='F1')
            PageTemplate.__init__(self, "DrawingPageTemplate", [content])

        def beforeDrawPage(self, canvas, doc) :
            """Draws a logo and an contribution message on each page."""
            someText = "Drawing generated image"
            canvas.saveState()
            canvas.setFont('Times-Roman', 10)
            canvas.drawCentredString(inch + (doc.pagesize[0] - (1.5 * inch)) / 2, 0.25 * inch, someText)
            canvas.restoreState()

    def __init__(self, context, filename, drawing) :
        self.context = context
        self.built = 0
        self.objects = []

        self.report = cStringIO.StringIO()

        #self.document = BaseDocTemplate(self.report)

        #self.document.addPageTemplates(self.DrawingPageTemplate(self))

        #self.StyleSheet = styles.getSampleStyleSheet()

        #self.append(drawing)

        #self.document.build(self.objects)

# Commented out the above lines and put in the next lines, according to this email:
# ---------------------------------------------------------------------------
# http://two.pairlist.net/pipermail/reportlab-users/2004-November/003523.html
#
# The following code fails with "LayoutError: Flowable <Paragraph at
# ...>test test test test test test  too large on page 2":
# 
# from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame,
# Paragraph
# from reportlab.lib.styles import ParagraphStyle
# from reportlab.lib.pagesizes import A4
# 
# doc = BaseDocTemplate('test_flowables.pdf')
# pt = PageTemplate(
#          frames=[Frame(x1=10, y1=10, width=A4[0]-20, height=A4[1]-20)],
#                  id='normal')
#
# doc.addPageTemplates(pt)
# doc.build([Paragraph('test '*200, ParagraphStyle('normal'))]*50)
# 
# I've discovered that attribute _postponed is set to avoid infinite loops
# with comment "this ought to be cleared when they are finally drawn!",
# but it's realy is cleared only when the whole story is passed. 
# Certainly, I can change the last line to "doc.build([Paragraph('test
# '*200, ParagraphStyle('normal')) for i in range(50)])" to work around
# this bug, but reusing of some flowable elements is quite common. 
# 
# I know I can wrap all reused flowables so that this attribute is set to
# unique wrapper instance and not shared flowable itself, but this would
# be annoying.  Any other ideas how to quick-fix it?
# 
# -- 
# Denis S. Otkidach
# http://www.python.ru/
# ---------------------------------------------------------------------------
# 
        doc = BaseDocTemplate(self.report)
        pt = PageTemplate(frames=[Frame(x1=10, y1=10, width=A4[0]-20, height=A4[1]-20)], id='normal')
	doc.addPageTemplates(pt)
	doc.build([Paragraph('Unfortunately, this is not the time scale :). '*200, ParagraphStyle('normal')) for i in range(50)])

        self.built = 1

    def __str__(self) :
        """Returns the PDF document as a string of text, or None if it's not ready yet."""
        if self.built :
            return self.report.getvalue()
        else :
            return None

    def append(self, object) :
        """Appends an object to our platypus "story" (using ReportLab's terminology)."""
        self.objects.append(object)

    def escapexml(self, s) :
        """Escape some xml entities."""
        s = string.strip(s)
        s = string.replace(s, "&", "&amp;")
        s = string.replace(s, "<", "&lt;")
        return string.replace(s, ">", "&gt;")

drawing_base_schema = BaseFolderSchema +  Schema((
     StringField('id',
                 required=0, ## Still actually required, but
                             ## the widget will supply the missing value
                             ## on non-submits
                 mode="rw",
                 accessor="getId",
                 mutator="setId",
                 default=None,
                 widget=IdWidget(label="Short Name",
                                 description="Should not contain spaces, underscores or mixed case. "\
                                 "Short Name is part of the item's web address.",
                                 visible={'view' : 'visible'}),
                 ),
    StringField('title',
                required=1,
                searchable=1,
                default='',
                accessor='Title',
                widget=StringWidget(label="Title",
                                    description="This name shows in the navigation menu."),
                ),
    StringField('description',
                isMetadata=1,
                accessor='Description',
                widget=TextAreaWidget(label='Description', 
		                      description='This description is optional, but not a bad idea for search text purposes.'),
                ),
    StringField('image_types',
                vocabulary=IMAGE_TYPES,
                mutator="setImageTypes",
                accessor="ImageTypes",
                widget=MultiSelectionWidget(label='Image Types',
                                            description='Select image types to generate.',
					    format='checkbox'),
                ),
    IntegerField('width',
                 required=1,
                 default='1000',
		 validators=("isInt"),
                 accessor='Width',
                 widget=IntegerWidget(label="Width",
                                      description="Width of Image, in pixels."),
                 ),
    IntegerField('height',
                 required=1,
                 default='4700',
		 validators=("isInt"),
                 accessor='Height',
                 widget=IntegerWidget(label="Height",
                                      description="Height of Image, in pixels."),
                 ),
    IntegerField('x_offset',
                 required=1,
                 default='25',
		 validators=("isInt"),
                 accessor='XOffset',
                 widget=IntegerWidget(label="X Offset",
                                      description="Spacing from left margin, in pixels."),
                 ),
    IntegerField('y_offset',
                 required=1,
                 default='35',
		 validators=("isInt"),
                 accessor='YOffset',
                 widget=IntegerWidget(label="Y Offset",
                                      description="Spacing from top margin, in pixels."),
                 ),
   BooleanField('scale_to_fit',
                 searchable = 0,
                 accessor='isScaleToFit',
                 widget=BooleanWidget(label='Scale to Fit',
                                      description='If yes, time scale will be scaled to fit in width x height size; if no, time scale will be cropped on right and bottom, if necessary.'),
                 ),
        ))

geologic_time_scale_schema = Schema((
    FloatField('end_time',
                 required=1,
                 default='0.0',
                 accessor='EndTime',
		 validators=("isDecimal"),
                 widget=DecimalWidget(label="End Time",
                                     description="End (Most Recent, or Top) Time for Time Scale, in millions of years."),
                 ),
    FloatField('start_time',
                 required=1,
                 default='4567.2',
                 accessor='StartTime',
		 validators=("isDecimal"),
                 widget=DecimalWidget(label="Start Time",
                                     description="Start (Earliest, or Bottom) Time for Time Scale, in millions of years."),
                 ),
   BooleanField('simple_version',
                 searchable = 0,
                 accessor='isSimpleVersion',
                 widget=BooleanWidget(label='Create Simple Version?',
                                      description='If yes, time scale will not include series within Paleozoic and Mesozoic.'),
                 ),
    LinesField('point_events_field',
               required=0,
	       default='\n'.join(sample_point_events),
               accessor="PointEvents",
               validators=('validate_point_events',),
               widget=LinesWidget(label='Point Events',
                                  description='Point Events (for a point in time), one per line, with these fields: label placement (t,b,l,r,c for top, bottom, left, right, or center), x position (in pixels), label angle (0 for horizontal, 90 for vertical), symbol type (only circle, for now), width of symbol(in pixels), height of symbol (in pixels), time (millions of years ago), text for label, red (0-255), green (0-255), blue (0-255)',),
               ),
    LinesField('range_events_field',
               required=0,
	       default='\n'.join(sample_range_events),
               accessor="RangeEvents",
               validators=('validate_range_events',),
               widget=LinesWidget(label='Range Events',
                                  description='Range Events (for a span of time), one per line, with these fields: label placement (t,b,l,r,c for top, bottom, left, right, or center), x position (in pixels), label angle (0 for horizontal, 90 for vertical), width of bar (in pixels), start time (millions of years ago), end time (millions of years ago), text for label, red (0-255), green (0-255), blue (0-255)',),
               ),
        ))

class GeologicTimeScale(BaseFolder):
    """
    A folder object to hold PNG, SVG, PDF and other images
    generated for a drawing type.
    """
    allowed_content_types = ('File')

    filter_content_types = 1

    global_allow = 1

    schema = drawing_base_schema + geologic_time_scale_schema

    content_icon = 'timescaleicon.gif'

    security = ClassSecurityInfo()

    actions = ({'id': 'view',
                'name': 'View',
                'action': 'string:${object_url}/GeologicTimeScale_view',
                'permissions': (CMFCorePermissions.View,)},
               {'id': 'references',
                'name': 'References',
                'action': 'string:${object_url}/reference_edit',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0},
               {'id': 'metadata',
                'name': 'Properties',
                'action': 'string:${object_url}/base_metadata',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0})

    def gatherInput(self):
        image_types = self.ImageTypes()
        height = self.Height()
	width = self.Width()
        x_offset = self.XOffset()
        y_offset = self.YOffset()
	end_time = self.EndTime()
	start_time = self.StartTime()
	scale_to_fit = self.isScaleToFit()
	simple_version = self.isSimpleVersion()
        point_events_input = self.PointEvents()
        range_events_input = self.RangeEvents()

        return (image_types, 
	        height, width, 
		x_offset, y_offset, 
		end_time, start_time, 
		scale_to_fit, 
		simple_version, 
		point_events_input, 
		range_events_input)
            
    # processForm is called for creation
    def processForm(self, data=1, metadata=0, REQUEST=None, values=None):
	ret = BaseFolder.processForm(self, data, metadata, REQUEST, values)

        self.createDrawing(self.gatherInput())

        return ret

    # update is called for edit
    def update(self, **kwargs):
        ret = BaseFolder.update(self, **kwargs)

        self.createDrawing(self.gatherInput())

        return ret

    def createDrawing(self, input_items):
        image_types_with_messages, height, width, x_offset, y_offset, end_time, start_time, scale_to_fit, simple_version, point_events_input, range_events_input = input_items

	# trim off the parenthetic messages in the image_types
	image_types = []
	for image_type in image_types_with_messages:
	    if len(image_type) > 0: # extra one returned
	        image_types.append(image_type.split()[0])

        time_scale = GeologicTimeScaleDrawing(image_types, 
	                                      height, width, 
				              x_offset, y_offset, 
					      end_time, start_time, 
					      scale_to_fit, 
					      simple_version, 
					      point_events_input, 
					      range_events_input)

	drawing = time_scale.getDrawing()

        # save the drawing in the selected formats
	if 'PNG' in image_types:
            existing_file = getattr(self, 'drawing.png', None)
            if existing_file:
                 self.manage_delObjects('drawing.png')

            from reportlab.graphics import renderPM
            render_result = renderPM.drawToString(drawing, fmt='PNG')
            self.invokeFactory('File', id='drawing.png')
            png_file = self['drawing.png']
            png_file.update_data(data=render_result)
    
	if 'JPG' in image_types:
            existing_file = getattr(self, 'drawing.jpg', None)
            if existing_file:
                 self.manage_delObjects('drawing.jpg')

            from reportlab.graphics import renderPM
            render_result = renderPM.drawToString(drawing, fmt='JPG')
            self.invokeFactory('File', id='drawing.jpg')
            jpg_file = self['drawing.jpg']
            jpg_file.update_data(data=render_result)

	if 'GIF' in image_types:
            existing_file = getattr(self, 'drawing.gif', None)
            if existing_file:
                 self.manage_delObjects('drawing.gif')

            from reportlab.graphics import renderPM
            render_result = renderPM.drawToString(drawing, fmt='GIF')
            self.invokeFactory('File', id='drawing.gif')
            gif_file = self['drawing.gif']
            gif_file.update_data(data=render_result)
    
	if 'TIFF' in image_types:
            existing_file = getattr(self, 'drawing.tiff', None)
            if existing_file:
                 self.manage_delObjects('drawing.tiff')

            from reportlab.graphics import renderPM
            render_result = renderPM.drawToString(drawing, fmt='TIFF')
            self.invokeFactory('File', id='drawing.tiff')
            tiff_file = self['drawing.tiff']
            tiff_file.update_data(data=render_result)
    
	if 'PCT' in image_types:
            existing_file = getattr(self, 'drawing.pct', None)
            if existing_file:
                 self.manage_delObjects('drawing.pct')

            from reportlab.graphics import renderPM
            render_result = renderPM.drawToString(drawing, fmt='PCT')
            self.invokeFactory('File', id='drawing.pct')
            pct_file = self['drawing.pct']
            pct_file.update_data(data=render_result)
    
        if 'SVG-via-ReportLab' in image_types:
            existing_file = getattr(self, 'drawing-via-ReportLab.html', None)
            if existing_file:
                 self.manage_delObjects('drawing-via-ReportLab.html')

            self.invokeFactory('File', id='drawing-via-ReportLab.html')
            svg_file = self['drawing-via-ReportLab.html']
            svg_file.update_data(data=''.join(time_scale.getHTML('drawing-via-ReportLab')))

            existing_file = getattr(self, 'drawing-via-ReportLab.svg', None)
            if existing_file:
                 self.manage_delObjects('drawing-via-ReportLab.svg')

            from reportlab.graphics import renderSVG
            render_result = renderSVG.drawToString(drawing)
            self.invokeFactory('File', id='drawing-via-ReportLab.svg')
            svg_file = self['drawing-via-ReportLab.svg']
            svg_file.update_data(data=render_result)

        if 'SVG-directly-emitted' in image_types:
            existing_file = getattr(self, 'drawing-directly-emitted.html', None)
            if existing_file:
                 self.manage_delObjects('drawing-directly-emitted.html')

            self.invokeFactory('File', id='drawing-directly-emitted.html')
            svg_file = self['drawing-directly-emitted.html']
            svg_file.update_data(data=''.join(time_scale.getHTML('drawing-directly-emitted')))

            existing_file = getattr(self, 'drawing-directly-emitted.svg', None)
            if existing_file:
                 self.manage_delObjects('drawing-directly-emitted.svg')

            self.invokeFactory('File', id='drawing-directly-emitted.svg')
            svg_file = self['drawing-directly-emitted.svg']
            svg_file.update_data(data=''.join(time_scale.getSVG()))
	    
        #     self.invokeFactory('File', id='svgecmascript.js')
        #     ecmascript_file = self['svgecmascript.js']
        #     ecmascript_string = string.join(drawing.getEcmascript())
        #ecmascript_file.file_edit(file=dummy.File('svgecmascript.js', data=string.join(drawing.getEcmascript())))
        #     ecmascript_file.update_data(data=ecmascript_string)

#    
#        PDF generation doesn't work because DrawingPDFDoc
#        uses the platypus page layout engine, and the page
#        size is a standard size, and treats the drawing as
#        a flowable, which dies because it is too large.
#        So, need to do this without using platypus, or
#        learn how to scale the page size to the drawing.
#
#        This might be a bug in ReportLab, but the page
#        size and scaling need to be set appropriately anyway.
#
        if 'PDF' in image_types:
            existing_file = getattr(self, 'drawing.pdf', None)
            if existing_file:
                 self.manage_delObjects('drawing.pdf')

            pdf_content = str(DrawingPDFDoc(self, 'drawing.pdf', drawing))
            self.invokeFactory('File', id='drawing.pdf')
            pdf_file = self['drawing.pdf']
            pdf_file.update_data(data=pdf_content)

        if 'PS' in image_types:
            existing_file = getattr(self, 'drawing.ps', None)
            if existing_file:
                 self.manage_delObjects('drawing.ps')

            from reportlab.graphics import renderPS
            render_result = renderPS.drawToString(drawing)
            self.invokeFactory('File', id='drawing.ps')
            ps_file = self['drawing.ps']
            ps_file.update_data(data=render_result)

registerType(GeologicTimeScale, PROJECTNAME)

